/*
 * NewSimpleNavigator.h
 *
 *  Created on: 17.09.2015
 *      Author: swenzel
 */

#ifndef NAVIGATION_NEWSIMPLENAVIGATOR_H_
#define NAVIGATION_NEWSIMPLENAVIGATOR_H_

#include "VNavigator.h"
#include "SimpleSafetyEstimator.h"

namespace vecgeom {
inline namespace VECGEOM_IMPL_NAMESPACE {

// A very basic implementation of a navigator ( brute force which scales linearly with the number of daughters )
template <bool MotherIsConvex = false>
class NewSimpleNavigator : public VNavigatorHelper<class NewSimpleNavigator<MotherIsConvex>, MotherIsConvex> {

private:
  VECGEOM_CUDA_HEADER_DEVICE
  NewSimpleNavigator()
      : VNavigatorHelper<class NewSimpleNavigator<MotherIsConvex>, MotherIsConvex>(SimpleSafetyEstimator::Instance()) {
  } VECGEOM_CUDA_HEADER_DEVICE virtual ~NewSimpleNavigator() {}

public:
  VECGEOM_FORCE_INLINE
  virtual Precision ComputeStepAndHittingBoundaryForLocalPoint(Vector3D<Precision> const &localpoint,
                                                               Vector3D<Precision> const &localdir,
                                                               Precision step_limit, NavigationState const &in_state,
                                                               NavigationState &out_state) const override
  {

    auto currentvolume = in_state.Top();
#if defined(VERBOSE) && !defined(VECGEOM_NVCC)
    __thread static size_t counter = 0;
    if (counter % 1 == 0) {
      std::cerr << GetClassName() << " navigating in " << currentvolume->GetLabel() << " stepnumber " << counter
                << "step_limit" << step_limit << "localpoint " << localpoint << "localdir" << globaldir;
    }
    counter++;
#endif

    Precision step;
    assert(currentvolume != nullptr && "currentvolume is null in navigation");
    step = currentvolume->DistanceToOut(localpoint, localdir, step_limit);
    // NOTE: IF STEP IS NEGATIVE HERE, SOMETHING IS TERRIBLY WRONG. SET TO INFINITY SO THAT WE CAN BETTER HANDLE IT
    // LATER ON
    if (step < 0.) {
      step = kInfinity;
    }

    // iterate over all daughters
    auto *daughters                       = currentvolume->GetLogicalVolume()->GetDaughtersp();
    decltype(currentvolume) nexthitvolume = nullptr;
    auto ndaughters                       = daughters->size();
    for (decltype(ndaughters) d = 0; d < ndaughters; ++d) {
      auto daughter = daughters->operator[](d);
//    previous distance becomes step estimate, distance to daughter returned in workspace
// SW: this makes the navigation more robust and it appears that I have to
// put this at the moment since not all shapes respond yet with a negative distance if
// the point is actually inside the daughter
#ifdef CHECKCONTAINS
      bool contains = daughter->Contains(localpoint);
      if (!contains) {
#endif
        Precision ddistance = daughter->DistanceToIn(localpoint, localdir, step);

        // if distance is negative; we are inside that daughter and should relocate
        // unless distance is minus infinity
        bool valid    = (ddistance < step && !IsInf(ddistance));
        nexthitvolume = valid ? daughter : nexthitvolume;
        step          = valid ? ddistance : step;
#ifdef CHECKCONTAINS
      } else {
        std::cerr << " INDA ";
        step          = -1.;
        nexthitvolume = daughter;
        break;
      }
#endif
    }
    bool done;
    return VNavigator::PrepareOutState(in_state, out_state, step, step_limit, nexthitvolume, done);
  }

  VECGEOM_FORCE_INLINE
  VECGEOM_CUDA_HEADER_BOTH
  virtual bool CheckDaughterIntersections(LogicalVolume const *lvol, Vector3D<Precision> const &localpoint,
                                          Vector3D<Precision> const &localdir, NavigationState const & /*in_state*/,
                                          NavigationState & /*out_state*/, Precision &step,
                                          VPlacedVolume const *&hitcandidate) const override
  {

    // iterate over all daughters
    auto *daughters = lvol->GetDaughtersp();
    auto ndaughters = daughters->size();
    for (decltype(ndaughters) d = 0; d < ndaughters; ++d) {
      auto daughter = daughters->operator[](d);
//    previous distance becomes step estimate, distance to daughter returned in workspace
// SW: this makes the navigation more robust and it appears that I have to
// put this at the moment since not all shapes respond yet with a negative distance if
// the point is actually inside the daughter
#ifdef CHECKCONTAINS
      bool contains = daughter->Contains(localpoint);
      if (!contains) {
#endif
        Precision ddistance = daughter->DistanceToIn(localpoint, localdir, step);

        // if distance is negative; we are inside that daughter and should relocate
        // unless distance is minus infinity
        bool valid   = (ddistance < step && !IsInf(ddistance));
        hitcandidate = valid ? daughter : hitcandidate;
        step         = valid ? ddistance : step;
#ifdef CHECKCONTAINS
      } else {
        std::cerr << " INDA ";
        step         = -1.;
        hitcandidate = daughter;
        break;
      }
#endif
    }
    return false;
  }

  // Vector specialization for NewSimpleNavigator
  // TODO: unify with scalar use case
  template <typename T, unsigned int ChunkSize> // we may go to Backend as template parameter in future
  static void DaughterIntersectionsLooper(
      VNavigator const * /*nav*/, LogicalVolume const *lvol, Vector3D<T> const &localpoint, Vector3D<T> const &localdir,
      //                                          NavStatePool const &in_states, NavStatePool &out_states,
      NavigationState const ** /*in_states*/, NavigationState ** /*out_states*/, unsigned int from_index,
      Precision *out_steps, VPlacedVolume const *hitcandidates[ChunkSize])
  {
    // dispatch to vector implementation
    // iterate over all daughters
    // using Real_v = typename Backend::Real_v;
    // using Bool_v = Real_v::Mask;
    T step(vecCore::FromPtr<T>(out_steps + from_index));
    const auto *daughters = lvol->GetDaughtersp();
    auto ndaughters       = daughters->size();
    for (decltype(ndaughters) d = 0; d < ndaughters; ++d) {
      auto daughter = daughters->operator[](d);
      // SW: this makes the navigation more robust and it appears that I have to
      // put this at the moment since not all shapes respond yet with a negative distance if
      // the point is actually inside the daughter
      //  #ifdef CHECKCONTAINS
      //        Bool_v contains = daughter->Contains(localpoint);
      //        if (!contains) {
      //  #endif
      T ddistance = daughter->DistanceToIn(localpoint, localdir, step);

      // if distance is negative; we are inside that daughter and should relocate
      // unless distance is minus infinity
      auto valid = (ddistance < step && !IsInf(ddistance));

      // serial treatment of hit candidate until I find a better way
      // we might do this using a cast to a double vector with subsequent masked assignment
      if (!vecCore::MaskEmpty(valid)) {
        for (unsigned int i = 0 /*valid.firstOne()*/; i < ChunkSize; ++i) {
          hitcandidates[i] = vecCore::MaskLaneAt(valid, i) ? daughter : hitcandidates[i];
        }

        vecCore::MaskedAssign(step, valid, ddistance);
        //  #ifdef CHECKCONTAINS
        //        } else {
        //          std::cerr << " INDA ";
        //          step = -1.;
        //          hitcandidate = daughter;
        //          break;
        //        }
        //  #endif
      }
    }
    vecCore::Store(step, out_steps + from_index);
  }

//  template <typename Backend>
//  VECGEOM_FORCE_INLINE
//  virtual Bool_v
//  CheckDaughterIntersectionsT(LogicalVolume const *lvol, Vector3D<typename Backend::Real_v> const &localpoint,
//                              Vector3D<typename Backend::Real_v> const &localdir, NavStatePool const &in_state,
//                              NavStatePool &out_state, unsigned int from_index, unsigned int to_index,
//                              typename Backend::Real_v &step, VPlacedVolume const **&hitcandidates) const override {
//
//    // iterate over all daughters
//    using Real_v = typename Backend::Real_v;
//    using Bool_v = Real_v::Mask;
//    const auto *daughters = lvol->GetDaughtersp();
//    const auto ndaughters = daughters->size();
//    for (decltype(ndaughters) d = 0; d < ndaughters; ++d) {
//      auto daughter = daughters->operator[](d);
//      // SW: this makes the navigation more robust and it appears that I have to
//      // put this at the moment since not all shapes respond yet with a negative distance if
//      // the point is actually inside the daughter
//      //  #ifdef CHECKCONTAINS
//      //        Bool_v contains = daughter->Contains(localpoint);
//      //        if (!contains) {
//      //  #endif
//      Real_v ddistance = daughter->DistanceToIn(localpoint, localdir, step);
//
//      // if distance is negative; we are inside that daughter and should relocate
//      // unless distance is minus infinity
//      Bool_v valid = (ddistance < step && !IsInf(ddistance));
//      // serial treatment of hit candidate until I find a better way
//      for (unsigned int index = valid.firstOne(); index < Real_v::Size; ++index) {
//        hitcandidates[i] = valid[i] ? daughter : hitcandidates[i];
//      }
//      step(valid) = ddistance; // or maskedAssign
//                               //  #ifdef CHECKCONTAINS
//                               //        } else {
//                               //          std::cerr << " INDA ";
//                               //          step = -1.;
//                               //          hitcandidate = daughter;
//                               //          break;
//                               //        }
//                               //  #endif
//    }
//    return false;
//  }

#ifndef VECGEOM_NVCC
  static VNavigator *Instance()
  {
    static NewSimpleNavigator instance;
    return &instance;
  }
#else
  VECGEOM_CUDA_HEADER_DEVICE
  static VNavigator *Instance();
#endif

  static constexpr const char *gClassNameString = "NewSimpleNavigator";
  typedef SimpleSafetyEstimator SafetyEstimator_t;
}; // end of class
}
} // end namespace

#endif /* NAVIGATION_NEWSIMPLENAVIGATOR_H_ */
